import pytest
import pulp
from .core import Problem, Variable, negate, logical_and, logical_or, minimum, maximum, logical_xor, implies
from .errors import NonBinaryVariableError, CitrusError, assert_binary

def test_that_negate_produces_negated_variable():
    p = Problem('negation test', pulp.LpMinimize)
    x = p.make_var('x', cat=pulp.LpBinary)
    y = negate(x)
    z = negate(y)
    p.setObjective(2 * x + y)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert x.value() == 0
    assert y.value() == 1
    assert z.value() == 0

def test_that_logical_and_produces_constrained_value():
    p = Problem('logical_and test', pulp.LpMinimize)
    t1 = p.make_var('t1', cat=pulp.LpBinary)
    t2 = p.make_var('t2', cat=pulp.LpBinary)
    f1 = p.make_var('f1', cat=pulp.LpBinary)
    f2 = p.make_var('f2', cat=pulp.LpBinary)

    tt = logical_and(t1, t2)
    tf = logical_and(t1, f1)
    ft = logical_and(f1, t1)
    ff = logical_and(f1, f2)
    p.addConstraint(t1 == 1)
    p.addConstraint(t2 == 1)
    p.addConstraint(f1 == 0)
    p.addConstraint(f2 == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tt.value() == 1
    assert tf.value() == 0
    assert ft.value() == 0
    assert ff.value() == 0

def test_that_anding_with_itself_is_okay():
    p = Problem('anding with self', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)

    tt = logical_and(t, t)
    tf = logical_and(t, f)
    ft = logical_and(f, t)
    ff = logical_and(f, f)
    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tt.value() == 1
    assert tf.value() == 0
    assert ft.value() == 0
    assert ff.value() == 0

def test_that_logical_or_produces_constrained_value():
    p = Problem('logical_or tests', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)

    tt = logical_or(t, t)
    tf = logical_or(t, f)
    ft = logical_or(f, t)
    ff = logical_or(f, f)
    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tt.value() == 1
    assert tf.value() == 1
    assert ft.value() == 1
    assert ff.value() == 0

def test_that_funcs_throws_on_non_binary_variable():
    p = Problem('problem', pulp.LpMinimize)
    x = p.make_var('x', cat=pulp.LpInteger)
    y = p.make_var('y', cat=pulp.LpInteger)

    with pytest.raises(NonBinaryVariableError):
        logical_and(x, y)
    with pytest.raises(NonBinaryVariableError):
        logical_or(x, y)
    with pytest.raises(NonBinaryVariableError):
        logical_xor(x, y)
    with pytest.raises(NonBinaryVariableError):
        implies(x, y)

def test_that_vars_from_diff_problems_raise_error():
    a = Problem('problem a', pulp.LpMinimize)
    b = Problem('problem b', pulp.LpMinimize)
    x = a.make_var('x', cat=pulp.LpBinary)
    y = b.make_var('y', cat=pulp.LpBinary)

    with pytest.raises(CitrusError):
        logical_or(x, y)
    with pytest.raises(CitrusError):
        logical_and(x, y)
    with pytest.raises(CitrusError):
        logical_xor(x, y)
    with pytest.raises(CitrusError):
        implies(x, y)

def test_that_from_lp_var_works():
    p = Problem('anding with self', pulp.LpMinimize)
    t = pulp.LpVariable('t', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)
    t = Variable.from_lp_var(t, p)

    tf = logical_and(t, f)
    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tf.value() == 0

def test_more_than_two_args_or():
    p = Problem('logical_or tests', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    t2 = p.make_var('t2', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)
    f2 = p.make_var('f2', cat=pulp.LpBinary)

    ttt = logical_or(t, t, t2)
    tff = logical_or(t, f, f2)
    ftf = logical_or(f, t, f2)
    ff2f = logical_or(f, f, f2)

    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.addConstraint(t2 == 1)
    p.addConstraint(f2 == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert ttt.value() == 1
    assert tff.value() == 1
    assert ftf.value() == 1
    assert ff2f.value() == 0

def test_more_than_two_args_and():
    p = Problem('logical_and tests', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    t2 = p.make_var('t2', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)
    f2 = p.make_var('f2', cat=pulp.LpBinary)

    ttt = logical_and(t, t, t2)
    tff = logical_and(t, f, f2)
    ftf = logical_and(f, t, f2)
    ff2f = logical_and(f, f, f2)

    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.addConstraint(t2 == 1)
    p.addConstraint(f2 == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert ttt.value() == 1
    assert tff.value() == 0
    assert ftf.value() == 0
    assert ff2f.value() == 0


def test_that_minimum_is_truly_min():
    p = Problem('minimum', pulp.LpMaximize)

    x = p.make_var('x', cat=pulp.LpContinuous)
    p.addConstraint(x <= 52)

    y = p.make_var('y', cat=pulp.LpContinuous)
    p.addConstraint(y <= 12)

    z = p.make_var('z', cat=pulp.LpContinuous)
    p.addConstraint(z <= 15)

    m = minimum(x, y, z)
    p.setObjective(x + y + z + m)

    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert m.value() == 12

def test_that_maximum_is_truly_max():
    p = Problem('maximum', pulp.LpMinimize)

    x = p.make_var('x', cat=pulp.LpContinuous)
    p.addConstraint(x >= 52)

    y = p.make_var('y', cat=pulp.LpContinuous)
    p.addConstraint(y >= 12)

    z = p.make_var('z', cat=pulp.LpContinuous)
    p.addConstraint(z >= 15)

    m = maximum(x, y, z)
    p.setObjective(x + y + z + m)

    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert m.value() == 52

def test_logical_xor():
    p = Problem('logical_xor tests', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)

    tt = logical_xor(t, t)
    tf = logical_xor(t, f)
    ft = logical_xor(f, t)
    ff = logical_xor(f, f)
    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tt.value() == 0
    assert tf.value() == 1
    assert ft.value() == 1
    assert ff.value() == 0

def test_implies():
    p = Problem('implies tests', pulp.LpMinimize)
    t = p.make_var('t', cat=pulp.LpBinary)
    f = p.make_var('f', cat=pulp.LpBinary)

    tt = implies(t, t)
    tf = implies(t, f)
    ft = implies(f, t)
    ff = implies(f, f)
    p.addConstraint(t == 1)
    p.addConstraint(f == 0)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert tt.value() == 1
    assert tf.value() == 0
    assert ft.value() == 1
    assert ff.value() == 1

def test_addition_retains_problem_reference():
    """
    adding two Variables should produce an AffineExpression
    (not an LpAffineExpression, which loses its reference to the Problem)
    """
    p = Problem('works with addition', pulp.LpMinimize)
    a = p.make_var('a', cat=pulp.LpContinuous)
    b = p.make_var('b', cat=pulp.LpContinuous)

    c = a + b
    assert c._problem == p, "c should retain the problem from a, b"

def test_maximum_operates_on_affine_expr():
    p = Problem('maximum test', pulp.LpMinimize)
    a = p.make_var('a', cat=pulp.LpContinuous)
    b = p.make_var('b', cat=pulp.LpContinuous)
    c = 2 * a + b

    p.addConstraint((a + b + c) <= 12, 'a + b + c<= 12')
    p.addConstraint(a >= 0, 'pos a')
    p.addConstraint(b >= 0, 'pos b')
    p.addConstraint(c >= 0, 'pos c')

    largest = maximum(a, b, c)
    p.setObjective(largest)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'

def test_constants_count_as_binary():
    p = Problem('constants binary', pulp.LpMinimize)
    a = p.make_var('a', cat=pulp.LpBinary)
    a.setInitialValue(1)
    a.fixValue()
    assert_binary(a)

def test_abs_value():
    p = Problem('abs value', pulp.LpMinimize)
    a = p.make_var('a', cat=pulp.LpContinuous)
    b = p.make_var('b', cat=pulp.LpContinuous)

    p.addConstraint(abs(a - b) <= 5, '|a - b| <= 5')
    p.addConstraint(a >= 10, 'a >= 10')
    p.addConstraint(abs(b) >= 0) # not active, just making sure we can apply abs to variables

    p.setObjective(b)
    p.solve()
    assert pulp.LpStatus[p.status] == 'Optimal'
    assert a.value() == 10
    assert b.value() == 5



from functools import reduce
import pulp
from .errors import assert_binary, assert_same_problem, MissingProblemReference


class Problem(pulp.LpProblem):
    def __init__(self, *args ,**kwargs):
        super().__init__(*args, **kwargs)
        self._synth_var_ix = 0

    def make_var(self, *args, **kwargs):
        return Variable(*args, **kwargs, problem=self)

    def _synth_var(self):
        name = str(self._synth_var_ix)
        self._synth_var_ix += 1
        return name

    def dicts(self, *args, **kwargs):
        ds = pulp.LpVariable.dicts(*args, **kwargs)
        return self._walk_dicts(ds)

    def _walk_dicts(self, ds):
        if isinstance(ds, pulp.LpVariable):
            return Variable.from_lp_var(ds, self)
        if isinstance(ds, dict):
            return { k: self._walk_dicts(v) for k, v in ds.items() }
        raise ValueError('Expected a dict or LpVariable. received {}'.format(ds))

class AffineExpression(pulp.LpAffineExpression):
    def __init__(self, *args, **kwargs):
        if 'problem' in kwargs:
            problem = kwargs.pop('problem')
        elif len(args) > 0 and isinstance(args[0], (Variable, AffineExpression)):
            problem = args[0]._problem
        else:
            raise MissingProblemReference('Cannot create AffineExpression without a reference to the problem')
        super().__init__(*args, **kwargs)
        self._problem = problem

    def __abs__(self):
        return abs_value(self)

    def copy(self):
        return AffineExpression(self)

    def emptyCopy(self):
        return AffineExpression(problem=self._problem)


class Variable(pulp.LpVariable):
    def __init__(self, *args, **kwargs):
        self._problem = kwargs.pop('problem')
        super().__init__(*args, **kwargs)

    @classmethod
    def from_lp_var(cls, lp_var: pulp.LpVariable, problem: Problem):
        lp_var.__class__ = cls
        lp_var._problem = problem
        return lp_var

    def __or__(self, other):
        return logical_or(self, other)

    def __and__(self, other):
        return logical_and(self, other)

    def __xor__(self, other):
        return logical_xor(self, other)

    def __abs__(self):
        return abs_value(self)

    # -- copied from LpVariable, but with a different constructor
    def __neg__(self):
        return - AffineExpression(self)

    def __add__(self, other):
        return AffineExpression(self) + other

    def __radd__(self, other):
        return AffineExpression(self) + other

    def __sub__(self, other):
        return AffineExpression(self) - other

    def __rsub__(self, other):
        return other - AffineExpression(self)

    def __mul__(self, other):
        return AffineExpression(self) * other

    def __rmul__(self, other):
        return AffineExpression(self) * other

    def __div__(self, other):
        return AffineExpression(self)/other

def abs_value(var: Variable) -> Variable:
    problem = var._problem
    z = problem.make_var(f'abs({var.name})_{problem._synth_var()}', cat=pulp.LpInteger)
    problem.addConstraint(var <= z, f'{var.name} <= abs({var.name}) _{problem._synth_var()}')
    problem.addConstraint(-z <= var, f'- abs({var.name}) <= {var.name} _{problem._synth_var()}')
    return z

def prefer_between(x: Variable, a: int, b: int) -> Variable:
    """
    Produces a variable that is
     - 0 when x lies within [a, b]
     - negative otherwise, and more negative with distance from [a, b]
    """
    return (b - a) - abs_value(x - a) - abs_value(x - b)


def negate(x: Variable):
    assert_binary(x)
    problem = x._problem
    y = problem.make_var('(NOT {})_{}'.format(x.name, problem._synth_var()), cat=pulp.LpBinary)
    problem.addConstraint(y == 1 - x)
    return y

def logical_and(*xs):
    """
    produce a variable that represents x && y

    That variable can then be used in constraints and the objective func.
    """
    if len(xs) == 1:
        return xs[0]

    assert_same_problem(*xs)
    for x in xs:
        assert_binary(x)

    model = x._problem

    name = '(' + '_AND_'.join(x.name for x in xs) + f')_{model._synth_var()}'
    z = model.make_var(name, cat=pulp.LpBinary)
    model.addConstraint(z >= pulp.lpSum(xs) - len(xs) + 1)
    for x in xs:
        model.addConstraint(z <= x)
    return z

def logical_or(*xs):
    if len(xs) == 1:
        return xs[0]

    assert_same_problem(*xs)
    for x in xs:
        assert_binary(x)

    model = x._problem

    name = '(' + '_OR_'.join(x.name for x in xs) + f')_{model._synth_var()}'
    z = model.make_var(name, cat=pulp.LpBinary)
    model.addConstraint(z <= pulp.lpSum(xs))
    for x in xs:
        model.addConstraint(z >= x)
    return z

def logical_xor(x: Variable, y: Variable):
    assert_same_problem(x, y)
    assert_binary(x)
    assert_binary(y)

    model = x._problem

    z = model.make_var('({} XOR {})_{}'.format(x.name, y.name, model._synth_var()), cat=pulp.LpBinary)
    model.addConstraint(z <= x + y)
    model.addConstraint(z >= x - y)
    model.addConstraint(z >= y - x)
    model.addConstraint(z <= 2 - x - y)

    return z

def implies(x: Variable, y: Variable):
    assert_same_problem(x, y)
    assert_binary(x)
    assert_binary(y)

    model = x._problem

    z = model.make_var('({} implies {})_{}'.format(x.name, y.name, model._synth_var()), cat=pulp.LpBinary)
    model.addConstraint(z <= 1 - x + y)
    model.addConstraint(z >= 1 - x)
    model.addConstraint(z >= y)
    return z


def minimum(*xs: Variable, name=None):
    if len(xs) == 1:
        return xs[0]
    assert_same_problem(*xs)
    model = xs[0]._problem
    m = model.make_var('{}_{}'.format(name or 'min', model._synth_var()), cat=pulp.LpContinuous)
    for x in xs:
        model.addConstraint(m <= x)
    return m

def maximum(*xs: Variable, name=None):
    if len(xs) == 1:
        return xs[0]
    assert_same_problem(*xs)
    model = xs[0]._problem
    m = model.make_var('{}_{}'.format(name or 'max', model._synth_var()), cat=pulp.LpContinuous)
    for x in xs:
        model.addConstraint(m >= x)
    return m



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['citrus']

package_data = \
{'': ['*']}

install_requires = \
['pulp>=2.0,<3.0']

setup_kwargs = {
    'name': 'citrus',
    'version': '0.0.5',
    'description': 'A more convenient interface to doing Binary Linear Programming with PuLP',
    'long_description': "# Citrus\n\nIntended to work like [PuLP](https://github.com/coin-or/pulp), but with a few convenience functions thrown in.\n\n```bash\npip install citrus\n```\n\n# Comparisons\n\n## ANDing two variables\n\n```python\n# without citrus\nimport pulp\np = pulp.LpProblem('and example', pulp.LpMinimize)\n\nx = pulp.LpVariable('x', cat=pulp.LpBinary)\ny = pulp.LpVariable('y', cat=pulp.LpBinary)\nx_and_y = pulp.LpVariable('x_and_y', cat=pulp.LpBinary)\nmodel.addConstraint(x_and_y >= x + y - 1)\nmodel.addConstraint(x_and_y <= x)\nmodel.addConstraint(x_and_y <= y)\n```\n\n```python\n# with citrus\nimport citrus\np = citrus.Problem('and example', pulp.LpMinimize)\n\nx = p.make_var('x', cat=pulp.LpBinary)\ny = p.make_var('y', cat=pulp.LpBinary)\nx_and_y = x & y\n# alternatively, x_and_y = citrus.logical_and(x, y)\n```\n\n## ORing two variables\n\n```python\n# without citrus\nimport pulp\np = pulp.LpProblem('or example', pulp.LpMinimize)\n\nx = pulp.LpVariable('x', cat=pulp.LpBinary)\ny = pulp.LpVariable('y', cat=pulp.LpBinary)\nx_or_y = pulp.LpVariable('x_or_y', cat=pulp.LpBinary)\nmodel.addConstraint(x_or_y <= x + y)\nmodel.addConstraint(x_or_y >= x)\nmodel.addConstraint(x_or_y >= y)\n```\n\n```python\n# with citrus\nimport citrus\np = citrus.Problem('or example', pulp.LpMinimize)\n\nx = p.make_var('x', cat=pulp.LpBinary)\ny = p.make_var('y', cat=pulp.LpBinary)\nx_or_y = x | y\n# alternatively, x_or_y = citrus.logical_or(x, y)\n```\n\n## Negating a variable\n\n```python\n# without citrus\np = pulp.LpProblem('negation test', pulp.LpMinimize)\n\nx = pulp.LpVariable('x', cat=pulp.LpBinary)\nnot_x = pulp.LpVariable('not_x', cat=pulp.LpBinary)\np.addConstraint(not_x == 1 - x)\n```\n\n```python\n# With citrus\nimport citrus\np = citrus.Problem('negation test', pulp.LpMinimize)\n\nx = p.make_var('x', cat=pulp.LpBinary)\nnot_x = citrus.negate(x)\n```\n\n# Tips & Tricks\n\nSometimes, you'll have many variables that you want to AND or OR together:\n\n```python\np = citrus.Problem('vacation at some point', pulp.Maximize)\n\nvacation_in_x_month = [\n  p.make_var('vacation in ' + month, cat=pulp.LpBinary)\n  for month in MONTHS\n]\n\ntake_a_vacation = reduce(citrus.logical_or, vacation_in_x_month)\np.addConstraint(take_a_vacation)\n```\n\n# API\n\n## Classes\n\n- `Variable` is a subclass of `pulp.LpVariable`. It adds the following methods:\n\n  - (classmethod) `from_lp_var`. Upgrade a `pulp.LpVariable` to a Variable.\n  - `__or__(self, other)` Compute the `logical_or` of two binary `Variable`s\n  - `__and__(self, other)` Compute the `logical_and` of two binary `Variable`s\n  - `__and__(self, other)` Compute the `logical_and` of two binary `Variable`s\n  - `__abs__(self)` Create a new Variable restricted to the absolute value of this one.\n\n- `Problem` A subclass of `pulp.LpProblem`. It adds the following method\n  - `make_var()` accepts the same arguments as `pulp.LpVariable`, but produces a `Variable`\n\n## Functions\n\n- `negate(x: Variable)` Produce a new `Variable` with the opposite value of `x`.\n- `logical_and(x: Variable, y: Variable)` Produce a new `Variable` constrained to take on the AND of `x` and `y`.\n- `logical_or(x: Variable, y: Variable)` Produce a new `Variable` constrained to take on the OR of `x` and `y`.\n- `logical_xor(x: Variable, y: Variable)` Produce a new `Variable` constrained to take on the XOR of `x` and `y`.\n- `implies(x: Variable, y: Variable)` Produce a new variable constrained to take on the value of `x => y`\n- `minimum(*xs)` Produce a new `Variable` that can be no larger than the smallest in `xs`\n- `maximum(*xy)` Produce a new `Variable` that can be no smaller than the largest in `xs`\n",
    'author': 'Brian Schiller',
    'author_email': 'bgschiller@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bgschiller/citrus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import pulp
from functools import reduce

class CitrusError(pulp.PulpError):
    pass

class NonBinaryVariableError(CitrusError):
    pass

class MissingProblemReference(CitrusError):
    pass

def assert_binary(var):
    if var.isBinary(): return
    if var.isConstant() and (var.value() == 1 or var.value() == 0): return
    raise NonBinaryVariableError(var.name)

def assert_same_problem_binary(x, y):
    if x._problem is not y._problem:
        raise CitrusError('Variables must be associated with the same problem.')
    return x # return for use in reduce(assert_same_problem, (x, y, z))


def assert_same_problem(*xs):
    return reduce(assert_same_problem_binary, xs)
